@@ -5,18 +5,20 @@
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public
  * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
+ * version 2 of the License, or (at your option) any later version.
  * 
  * This software is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * General Public License for more details.
  * 
- * You should have received a copy of the GNU General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation, Inc.,
+ * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
+#include "b_config.h"
+
 #include <stdio.h>
 #include <ctype.h>
 #include <stdlib.h>
@@ -27,7 +29,7 @@
 #include <grp.h>
 #include <errno.h>
 #include <string.h>
-#include "b_config.h"
+#include <netdb.h>
 #include "booth.h"
 #include "config.h"
 #include "raft.h"
@@ -60,9 +62,60 @@ static int ticket_realloc(void)
 	return 0;
 }
 
+static void hostname_to_ip(char * hostname)
+{
+	struct addrinfo hints;
+	struct addrinfo *result, *rp;
+	int res;
+	int addr_found = 0;
+	const char *ntop_res;
 
-int add_site(char *address, int type);
-int add_site(char *addr_string, int type)
+	memset(&hints, 0, sizeof(hints));
+	hints.ai_family = AF_UNSPEC;
+	hints.ai_socktype = SOCK_STREAM;
+
+	res = getaddrinfo(hostname, NULL, &hints, &result);
+
+	if (res != 0) {
+		log_error("can't find IP for the host \"%s\"", hostname);
+		return;
+	}
+
+	/* Return the first found AF_INET or AF_INET6 address */
+	for (rp = result; rp && !addr_found; rp = rp->ai_next) {
+		if (rp->ai_family != AF_INET && rp->ai_family != AF_INET6) {
+			continue ;
+		}
+
+		switch (rp->ai_family) {
+		case AF_INET:
+			ntop_res = inet_ntop(rp->ai_family,
+			    &((struct sockaddr_in *)(rp->ai_addr))->sin_addr,
+			    hostname, BOOTH_NAME_LEN - 1);
+			break;
+		case AF_INET6:
+			ntop_res = inet_ntop(rp->ai_family,
+			    &((struct sockaddr_in6 *)(rp->ai_addr))->sin6_addr,
+			    hostname, BOOTH_NAME_LEN - 1);
+			break;
+		}
+
+		if (ntop_res) {
+			/* buffer overflow will not happen (IPv6 notation < 63 chars),
+			   but suppress the warnings */
+			hostname[BOOTH_NAME_LEN - 1] = '\0';
+			addr_found = 1;
+		}
+	}
+
+	if (!addr_found) {
+		log_error("no IP addresses found for the host \"%s\"", hostname);
+	}
+
+	freeaddrinfo(result);
+}
+
+static int add_site(char *addr_string, int type)
 {
 	int rv;
 	struct booth_site *site;
@@ -70,13 +123,13 @@ int add_site(char *addr_string, int type)
 	uint32_t mask;
 	int i;
 
-
 	rv = 1;
 	if (booth_conf->site_count == MAX_NODES) {
 		log_error("too many nodes");
 		goto out;
 	}
-	if (strlen(addr_string)+1 >= sizeof(booth_conf->site[0].addr_string)) {
+	if (strnlen(addr_string, sizeof(booth_conf->site[0].addr_string))
+			>= sizeof(booth_conf->site[0].addr_string)) {
 		log_error("site address \"%s\" too long", addr_string);
 		goto out;
 	}
@@ -85,10 +138,20 @@ int add_site(char *addr_string, int type)
 
 	site->family = AF_INET;
 	site->type = type;
-	/* Make site_id start at a non-zero point.
-	 * Perhaps use hash over string or address? */
-	strcpy(site->addr_string, addr_string);
 
+	/* buffer overflow will not hapen (we've already checked that
+	   addr_string will fit incl. terminating '\0' above), but
+	   suppress the warnings with copying everything but the boundary
+	   byte, which is valid as-is, since this last byte will be safely
+	   pre-zeroed from the struct booth_config initialization */
+	strncpy(site->addr_string, addr_string, sizeof(site->addr_string) - 1);
+
+	if (!(inet_pton(AF_INET, site->addr_string, &site->sa4.sin_addr) > 0) &&
+        !(inet_pton(AF_INET6, site->addr_string, &site->sa6.sin6_addr) > 0)) {
+
+		/* Not a valid address, so let us try to convert it into an IP address */
+		hostname_to_ip(site->addr_string);
+	}
 
 	site->index = booth_conf->site_count;
 	site->bitmask = 1 << booth_conf->site_count;
@@ -112,7 +175,10 @@ int add_site(char *addr_string, int type)
 	 * increase even more.
 	 * Whether there'll be a collision in real-life, with 3 or 5 nodes, is
 	 * another question ... but for now get the ID from the binary
-	 * representation - that had *no* collisions up to 32.0.0.0. */
+	 * representation - that had *no* collisions up to 32.0.0.0.
+	 * Note that POSIX mandates inet_pton to arange the address pointed
+	 * to by "dst" in network byte order, assuring little/big-endianess
+	 * mutual compatibility. */
 	if (inet_pton(AF_INET,
 				site->addr_string,
 				&site->sa4.sin_addr) > 0) {
@@ -231,6 +297,7 @@ static int add_ticket(const char *name, struct ticket_config **tkp,
 	tk->term_duration = def->term_duration;
 	tk->retries = def->retries;
 	memcpy(tk->weight, def->weight, sizeof(tk->weight));
+	tk->mode = def->mode;
 
 	if (tkp)
 		*tkp = tk;
@@ -303,6 +370,15 @@ static int parse_weights(const char *input, int weights[MAX_NODES])
 	return i;
 }
 
+/* returns TICKET_MODE_AUTO if failed to parse the ticket mode. */
+static ticket_mode_e retrieve_ticket_mode(const char *input)
+{
+	if (strcasecmp(input, "manual") == 0) {
+		return TICKET_MODE_MANUAL;
+	}
+
+	return TICKET_MODE_AUTO;
+}
 
 /* scan val for time; time is [0-9]+(ms)?, i.e. either in seconds
  * or milliseconds
@@ -330,7 +406,7 @@ static long read_time(char *val)
 }
 
 /* make arguments for execv(2)
- * tk_test.prog points to the path
+ * tk_test.path points to the path
  * tk_test.argv is argument vector (starts with the prog)
  * (strtok pokes holes in the configuration parameter value, i.e.
  * we don't need to allocate memory for arguments)
@@ -340,21 +416,21 @@ static int parse_extprog(char *val, struct ticket_config *tk)
 	char *p;
 	int i = 0;
 
-	if (tk_test.prog) {
-		free(tk_test.prog);
+	if (tk_test.path) {
+		free(tk_test.path);
 	}
-	if (!(tk_test.prog = strdup(val))) {
+	if (!(tk_test.path = strdup(val))) {
 		log_error("out of memory");
 		return -1;
 	}
 
-	p = strtok(tk_test.prog, " \t");
+	p = strtok(tk_test.path, " \t");
 	tk_test.argv[i++] = p;
 	do {
 		p = strtok(NULL, " \t");
 		if (i >= MAX_ARGS) {
 			log_error("too many arguments for the acquire-handler");
-			free(tk_test.prog);
+			free(tk_test.path);
 			return -1;
 		}
 		tk_test.argv[i++] = p;
@@ -465,6 +541,7 @@ extern int poll_timeout;
 int read_config(const char *path, int type)
 {
 	char line[1024];
+	char error_str_buf[1024];
 	FILE *fp;
 	char *s, *key, *val, *end_of_key;
 	const char *error;
@@ -486,6 +563,7 @@ int read_config(const char *path, int type)
 	booth_conf = malloc(sizeof(struct booth_config)
 			+ TICKET_ALLOC * sizeof(struct ticket_config));
 	if (!booth_conf) {
+		fclose(fp);
 		log_error("failed to alloc memory for booth config");
 		return -ENOMEM;
 	}
@@ -509,7 +587,7 @@ int read_config(const char *path, int type)
 	strcpy(booth_conf->arb_group,  "nobody");
 
 	parse_weights("", defaults.weight);
-	defaults.clu_test.prog  = NULL;
+	defaults.clu_test.path  = NULL;
 	defaults.clu_test.pid  = 0;
 	defaults.clu_test.status  = 0;
 	defaults.clu_test.progstate  = EXTPROG_IDLE;
@@ -517,6 +595,7 @@ int read_config(const char *path, int type)
 	defaults.timeout       = DEFAULT_TICKET_TIMEOUT;
 	defaults.retries       = DEFAULT_RETRIES;
 	defaults.acquire_after = 0;
+	defaults.mode          = TICKET_MODE_AUTO;
 
 	error = "";
 
@@ -620,7 +699,9 @@ no_value:
 			else if (strcasecmp(val, "SCTP") == 0)
 				booth_conf->proto = SCTP;
 			else {
-				error = "invalid transport protocol";
+				(void)snprintf(error_str_buf, sizeof(error_str_buf),
+				    "invalid transport protocol \"%s\"", val);
+				error = error_str_buf;
 				goto err;
 			}
 			got_transport = 1;
@@ -655,13 +736,13 @@ no_value:
 
 		if (strcmp(key, "site") == 0) {
 			if (add_site(val, SITE))
-				goto out;
+				goto err;
 			continue;
 		}
 
 		if (strcmp(key, "arbitrator") == 0) {
 			if (add_site(val, ARBITRATOR))
-				goto out;
+				goto err;
 			continue;
 		}
 
@@ -695,13 +776,13 @@ no_value:
 		if (strcmp(key, "ticket") == 0) {
 			if (current_tk && strcmp(current_tk->name, "__defaults__")) {
 				if (!postproc_ticket(current_tk)) {
-					goto out;
+					goto err;
 				}
 			}
 			if (!strcmp(val, "__defaults__")) {
 				current_tk = &defaults;
 			} else if (add_ticket(val, &current_tk, &defaults)) {
-				goto out;
+				goto err;
 			}
 			continue;
 		}
@@ -710,7 +791,9 @@ no_value:
 		 * we don't know to which ticket the key refers
 		 */
 		if (!current_tk) {
-			error = "Unexpected keyword";
+			(void)snprintf(error_str_buf, sizeof(error_str_buf),
+			    "Unexpected keyword \"%s\"", key);
+			error = error_str_buf;
 			goto err;
 		}
 
@@ -779,15 +862,23 @@ no_value:
 			continue;
 		}
 
-		if (strcmp(key, "weights") == 0) {
-			if (parse_weights(val, current_tk->weight) < 0)
-				goto out;
+		if (strcmp(key, "mode") == 0) {
+			current_tk->mode = retrieve_ticket_mode(val);
 			continue;
 		}
 
-		error = "Unknown keyword";
+		if (strcmp(key, "weights") == 0) {
+			if (parse_weights(val, current_tk->weight) < 0)
+				goto err;
+			continue;
+		}
+
+		(void)snprintf(error_str_buf, sizeof(error_str_buf),
+		    "Unknown keyword \"%s\"", key);
+		error = error_str_buf;
 		goto err;
 	}
+	fclose(fp);
 
 	if ((booth_conf->site_count % 2) == 0) {
 		log_warn("Odd number of nodes is strongly recommended!");
@@ -802,7 +893,7 @@ no_value:
 			cp2 = cp + strlen(cp);
 		if (cp2-cp >= BOOTH_NAME_LEN) {
 			log_error("booth config file name too long");
-			goto err;
+			goto out;
 		}
 		strncpy(booth_conf->name, cp, cp2-cp);
 		*(booth_conf->name+(cp2-cp)) = '\0';
@@ -820,6 +911,7 @@ no_value:
 
 
 err:
+	fclose(fp);
 out:
 	log_error("%s in config file line %d",
 			error, lineno);
@@ -909,7 +1001,7 @@ static int get_other_site(struct booth_site **node)
 }
 
 
-int find_site_by_name(unsigned char *site, struct booth_site **node, int any_type)
+int find_site_by_name(char *site, struct booth_site **node, int any_type)
 {
 	struct booth_site *n;
 	int i;
@@ -923,7 +1015,7 @@ int find_site_by_name(unsigned char *site, struct booth_site **node, int any_typ
 	for (i = 0; i < booth_conf->site_count; i++) {
 		n = booth_conf->site + i;
 		if ((n->type == SITE || any_type) &&
-		    strcmp(n->addr_string, site) == 0) {
+		    strncmp(n->addr_string, site, sizeof(n->addr_string)) == 0) {
 			*node = n;
 			return 1;
 		}
