@@ -5,18 +5,20 @@
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public
  * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
+ * version 2 of the License, or (at your option) any later version.
  * 
  * This software is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * General Public License for more details.
  * 
- * You should have received a copy of the GNU General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation, Inc.,
+ * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
+#include "b_config.h"
+
 #include <stdio.h>
 #include <unistd.h>
 #include <stdlib.h>
@@ -33,22 +35,29 @@
 #include <sys/un.h>
 #include <sys/poll.h>
 #include <sys/wait.h>
-#include <pacemaker/crm/services.h>
-#include <clplumbing/setproctitle.h>
-#include <sys/prctl.h>
-#include <clplumbing/coredumps.h>
 #include <fcntl.h>
 #include <string.h>
 #include <ctype.h>
 #include <assert.h>
-#include <error.h>
-#include <sys/ioctl.h>
-#include <termios.h>
 #include <signal.h>
 #include <netdb.h>
 #include <arpa/inet.h>
 #include <sys/types.h>
-#include "b_config.h"
+
+#include <crm/services.h>
+
+#if HAVE_LIBGCRYPT
+#include <gcrypt.h>
+#endif
+#ifndef NAMETAG_LIBSYSTEMD
+#include <clplumbing/setproctitle.h>
+#else
+#include "alt/nametag_libsystemd.h"
+#endif
+#ifdef COREDUMP_NURSING
+#include <sys/prctl.h>
+#include <clplumbing/coredumps.h>
+#endif
 #include "log.h"
 #include "booth.h"
 #include "config.h"
@@ -58,13 +67,13 @@
 #include "ticket.h"
 #include "request.h"
 #include "attr.h"
+#include "handler.h"
 
-#define RELEASE_VERSION		"1.0"
-#define RELEASE_STR 	RELEASE_VERSION " (build " BOOTH_BUILD_VERSION ")"
+#define RELEASE_STR 	VERSION
 
 #define CLIENT_NALLOC		32
 
-int daemonize = 0;
+static int daemonize = 1;
 int enable_stderr = 0;
 timetype start_time;
 
@@ -83,8 +92,9 @@ static int client_size = 0;
 static const struct booth_site _no_leader = {
 	.addr_string = "none",
 	.site_id = NO_ONE,
+	.index = -1,
 };
-struct booth_site *no_leader = (struct booth_site*)& _no_leader;
+struct booth_site *const no_leader = (struct booth_site*) &_no_leader;
 
 typedef enum
 {
@@ -99,20 +109,23 @@ int poll_timeout;
 struct booth_config *booth_conf;
 struct command_line cl;
 
+/*
+ * Global signal handlers variables
+ */
+static int sig_exit_handler_called = 0;
+static int sig_exit_handler_sig = 0;
+static int sig_usr1_handler_called = 0;
+static int sig_chld_handler_called = 0;
+
 static void client_alloc(void)
 {
 	int i;
 
-	if (!clients) {
-		clients = malloc(CLIENT_NALLOC * sizeof(struct client));
-		pollfds = malloc(CLIENT_NALLOC * sizeof(struct pollfd));
-	} else {
-		clients = realloc(clients, (client_size + CLIENT_NALLOC) *
-					sizeof(struct client));
-		pollfds = realloc(pollfds, (client_size + CLIENT_NALLOC) *
-					sizeof(struct pollfd));
-	}
-	if (!clients || !pollfds) {
+	if (!(clients = realloc(
+		clients, (client_size + CLIENT_NALLOC) * sizeof(*clients))
+	) || !(pollfds = realloc(
+		pollfds, (client_size + CLIENT_NALLOC) * sizeof(*pollfds))
+	)) {
 		log_error("can't alloc for client array");
 		exit(1);
 	}
@@ -241,8 +254,10 @@ static int format_peers(char **pdata, unsigned int *len)
 				s->recv_err_cnt,
 				s->sec_cnt,
 				s->invalid_cnt);
-		if (alloc - (cp - data) <= 0)
+		if (alloc - (cp - data) <= 0) {
+			free(data);
 			return -ENOMEM;
+		}
 	}
 
 	*pdata = data;
@@ -255,7 +270,7 @@ static int format_peers(char **pdata, unsigned int *len)
 void list_peers(int fd)
 {
 	char *data;
-	int olen;
+	unsigned int olen;
 	struct boothc_hdr_msg hdr;
 
 	if (format_peers(&data, &olen) < 0)
@@ -273,7 +288,7 @@ out:
  */
 static void trim_key()
 {
-	unsigned char *p;
+	char *p;
 	int i;
 
 	for (i=0, p=booth_conf->authkey; i < booth_conf->authkey_len; i++, p++)
@@ -299,21 +314,24 @@ static int read_authkey()
 	int fd;
 
 	booth_conf->authkey[0] = '\0';
-	if (stat(booth_conf->authfile, &booth_conf->authstat) < 0) {
-		log_error("cannot stat authentication file %s: %s",
-			booth_conf->authfile, strerror(errno));
-		return -1;
-	}
-	if (booth_conf->authstat.st_mode & (S_IRGRP | S_IROTH)) {
-		log_error("%s: file can be readable only for the owner", booth_conf->authfile);
-		return -1;
-	}
 	fd = open(booth_conf->authfile, O_RDONLY);
 	if (fd < 0) {
 		log_error("cannot open %s: %s",
 			booth_conf->authfile, strerror(errno));
 		return -1;
 	}
+	if (fstat(fd, &booth_conf->authstat) < 0) {
+		log_error("cannot stat authentication file %s (%d): %s",
+			booth_conf->authfile, fd, strerror(errno));
+		close(fd);
+		return -1;
+	}
+	if (booth_conf->authstat.st_mode & (S_IRGRP | S_IROTH)) {
+		log_error("%s: file shall not be readable for anyone but the owner",
+			booth_conf->authfile);
+		close(fd);
+		return -1;
+	}
 	booth_conf->authkey_len = read(fd, booth_conf->authkey, BOOTH_MAX_KEY_LEN);
 	close(fd);
 	trim_key();
@@ -350,6 +368,15 @@ static int setup_config(int type)
 		rv = read_authkey();
 		if (rv < 0)
 			goto out;
+#if HAVE_LIBGCRYPT
+		if (!gcry_check_version(NULL)) {
+			log_error("gcry_check_version");
+			rv = -ENOENT;
+			goto out;
+		}
+		gcry_control(GCRYCTL_DISABLE_SECMEM, 0);
+		gcry_control(GCRYCTL_INITIALIZATION_FINISHED, 0);
+#endif
 	}
 
 	/* Set "local" pointer, ignoring errors. */
@@ -461,6 +488,24 @@ static int write_daemon_state(int fd, int state)
 	return 0;
 }
 
+static int process_signals(void)
+{
+	if (sig_exit_handler_called) {
+		log_info("caught signal %d", sig_exit_handler_sig);
+		return 1;
+	}
+	if (sig_usr1_handler_called) {
+		sig_usr1_handler_called = 0;
+		tickets_log_info();
+	}
+	if (sig_chld_handler_called) {
+		sig_chld_handler_called = 0;
+		wait_child(SIGCHLD);
+	}
+
+	return 0;
+}
+
 static int loop(int fd)
 {
 	void (*workfn) (int ci);
@@ -513,6 +558,10 @@ static int loop(int fd)
 		}
 
 		process_tickets();
+
+		if (process_signals() != 0) {
+			return 0;
+		}
 	}
 
 	return 0;
@@ -713,7 +762,7 @@ static int do_command(cmd_request_t cmd)
 	else if (cmd == CMD_REVOKE)
 		op_str = "revoke";
 
-	rv = 0;
+	rv = -1;
 	site = NULL;
 
 	/* Always use TCP for client - at least for now. */
@@ -745,7 +794,8 @@ static int do_command(cmd_request_t cmd)
 	if (!cl.msg.ticket.id[0]) {
 		/* If the loaded configuration has only a single ticket defined, use that. */
 		if (booth_conf->ticket_count == 1) {
-			strcpy(cl.msg.ticket.id, booth_conf->ticket[0].name);
+			strncpy(cl.msg.ticket.id, booth_conf->ticket[0].name,
+				sizeof(cl.msg.ticket.id));
 		} else {
 			log_error("No ticket given.");
 			goto out_close;
@@ -822,7 +872,7 @@ static int _lockfile(int mode, int *fdp, pid_t *locked_by)
 	 * Try to create it, but ignore errors. */
 	if (strncmp(cl.lockfile, BOOTH_RUN_DIR,
 				strlen(BOOTH_RUN_DIR)) == 0)
-		mkdir(BOOTH_RUN_DIR, 0775);
+		(void)mkdir(BOOTH_RUN_DIR, 0775);
 
 
 	if (locked_by)
@@ -924,6 +974,8 @@ static void print_usage(void)
 	"  -s <site>     Connect/grant to a different site\n"
 	"  -F            Try to grant the ticket immediately\n"
 	"                even if not all sites are reachable\n"
+	"                For manual tickets:\n"
+	"                grant a manual ticket even if it has been already granted\n"
 	"  -w            Wait forever for the outcome of the request\n"
 	"  -C            Wait until the ticket is committed to the CIB (grant only)\n"
 	"  -h            Print this help\n"
@@ -1098,12 +1150,14 @@ static int read_arguments(int argc, char **argv)
 					strcat(cp, BOOTH_DEFAULT_CONF_EXT);
 			}
 			break;
+
 		case 'D':
 			debug_level++;
-			enable_stderr = 1;
-			/* Fall through */
+			break;
+
 		case 'S':
-			daemonize = 1;
+			daemonize = 0;
+			enable_stderr = 1;
 			break;
 
 		case 'l':
@@ -1233,11 +1287,15 @@ static void set_scheduler(void)
 
 	rlimit.rlim_cur = RLIM_INFINITY;
 	rlimit.rlim_max = RLIM_INFINITY;
-	setrlimit(RLIMIT_MEMLOCK, &rlimit);
-	rv = mlockall(MCL_CURRENT | MCL_FUTURE);
+	rv = setrlimit(RLIMIT_MEMLOCK, &rlimit);
 	if (rv < 0) {
-		log_error("mlockall failed");
-	}
+		log_error("setrlimit failed");
+	} else {
+                rv = mlockall(MCL_CURRENT | MCL_FUTURE);
+                if (rv < 0) {
+                        log_error("mlockall failed");
+                }
+        }
 
 	rv = sched_get_priority_max(SCHED_RR);
 	if (rv != -1) {
@@ -1269,7 +1327,7 @@ static int set_procfs_val(const char *path, const char *val)
 static int do_status(int type)
 {
 	pid_t pid;
-	int rv, lock_fd, ret;
+	int rv, status_lock_fd, ret;
 	const char *reason = NULL;
 	char lockfile_data[1024], *cp;
 
@@ -1290,33 +1348,31 @@ static int do_status(int type)
 	}
 
 
-	rv = _lockfile(O_RDWR, &lock_fd, &pid);
-	if (rv == 0) {
-		reason = "PID file not locked.";
-		goto quit;
-	}
-	if (lock_fd == -1) {
+	rv = _lockfile(O_RDWR, &status_lock_fd, &pid);
+	if (status_lock_fd == -1) {
 		reason = "No PID file.";
 		goto quit;
 	}
-
+	if (rv == 0) {
+		close(status_lock_fd);
+		reason = "PID file not locked.";
+		goto quit;
+	}
 	if (pid) {
 		fprintf(stdout, "booth_lockpid=%d ", pid);
 		fflush(stdout);
 	}
 
-
-	rv = read(lock_fd, lockfile_data, sizeof(lockfile_data) - 1);
+	rv = read(status_lock_fd, lockfile_data, sizeof(lockfile_data) - 1);
 	if (rv < 4) {
+		close(status_lock_fd);
 		reason = "Cannot read lockfile data.";
 		ret = PCMK_LSB_UNKNOWN_ERROR;
 		goto quit;
-
 	}
 	lockfile_data[rv] = 0;
 
-	if (lock_fd != -1)
-		close(lock_fd);
+	close(status_lock_fd);
 
 
 	/* Make sure it's only a single line */
@@ -1338,7 +1394,7 @@ static int do_status(int type)
 
 	fprintf(stdout, "booth_lockfile='%s' %s\n",
 			cl.lockfile, lockfile_data);
-	if (daemonize)
+	if (!daemonize)
 		fprintf(stderr, "Booth at %s port %d seems to be running.\n",
 				local->addr_string, booth_conf->port);
 	return 0;
@@ -1347,7 +1403,7 @@ static int do_status(int type)
 quit:
 	log_debug("not running: %s", reason);
 	/* Ie. "DEBUG" */
-	if (daemonize)
+	if (!daemonize)
 		fprintf(stderr, "not running: %s\n", reason);
 	return ret;
 }
@@ -1392,33 +1448,18 @@ static void server_exit(void)
 
 static void sig_exit_handler(int sig)
 {
-	log_info("caught signal %d", sig);
-	exit(0);
+	sig_exit_handler_sig = sig;
+	sig_exit_handler_called = 1;
 }
 
-static void wait_child(int sig)
+static void sig_usr1_handler(int sig)
 {
-	int i, status;
-	struct ticket_config *tk;
+	sig_usr1_handler_called = 1;
+}
 
-	/* use waitpid(2) and not wait(2) in order not to interfear
-	 * with popen(2)/pclose(2) and system(2) used in pacemaker.c
-	 */
-	foreach_ticket(i, tk) {
-		if (tk_test.prog && tk_test.pid >= 0 &&
-				(tk_test.progstate == EXTPROG_RUNNING ||
-				tk_test.progstate == EXTPROG_IGNORE) &&
-				waitpid(tk_test.pid, &status, WNOHANG) == tk_test.pid) {
-			if (tk_test.progstate == EXTPROG_IGNORE) {
-				/* not interested in the outcome */
-				tk_test.pid = 0;
-				tk_test.progstate = EXTPROG_IDLE;
-			} else {
-				tk_test.status = status;
-				tk_test.progstate = EXTPROG_EXITED;
-			}
-		}
-	}
+static void sig_chld_handler(int sig)
+{
+	sig_chld_handler_called = 1;
 }
 
 static int do_server(int type)
@@ -1435,21 +1476,30 @@ static int do_server(int type)
 		exit(EXIT_FAILURE);
 	}
 
-	if (!daemonize) {
+	if (daemonize) {
 		if (daemon(0, 0) < 0) {
 			perror("daemon error");
 			exit(EXIT_FAILURE);
 		}
 	}
 
+	/*
+	 * Register signal and exit handler
+	 */
+	signal(SIGUSR1, (__sighandler_t)sig_usr1_handler);
+	signal(SIGTERM, (__sighandler_t)sig_exit_handler);
+	signal(SIGINT, (__sighandler_t)sig_exit_handler);
+	/* we'll handle errors there and then */
+	signal(SIGPIPE, SIG_IGN);
+
+	atexit(server_exit);
+
 	/* The lockfile must be written to _after_ the call to daemon(), so
 	 * that the lockfile contains the pid of the daemon, not the parent. */
 	lock_fd = create_lockfile();
 	if (lock_fd < 0)
 		return lock_fd;
 
-	atexit(server_exit);
-
 	strcat(log_ent, type_to_string(local->type));
 	cl_log_set_entity(log_ent);
 	cl_log_enable_stderr(enable_stderr ? TRUE : FALSE);
@@ -1459,11 +1509,6 @@ static int do_server(int type)
 	log_info("BOOTH %s %s daemon is starting",
 			type_to_string(local->type), RELEASE_STR);
 
-	signal(SIGUSR1, (__sighandler_t)tickets_log_info);
-	signal(SIGTERM, (__sighandler_t)sig_exit_handler);
-	signal(SIGINT, (__sighandler_t)sig_exit_handler);
-	/* we'll handle errors there and then */
-	signal(SIGPIPE, SIG_IGN);
 
 	set_scheduler();
 	/* we don't want to be killed by the OOM-killer */
@@ -1480,13 +1525,19 @@ static int do_server(int type)
 	if (rv)
 		return rv;
 
+#ifdef COREDUMP_NURSING
 	if (cl_enable_coredumps(TRUE) < 0){
-		cl_log(LOG_ERR, "enabling core dump failed");
+		log_error("enabling core dump failed");
 	}
 	cl_cdtocoredir();
 	prctl(PR_SET_DUMPABLE, (unsigned long)TRUE, 0UL, 0UL, 0UL);
+#else
+	if (chdir(BOOTH_CORE_DIR) < 0) {
+		log_error("cannot change working directory to %s", BOOTH_CORE_DIR);
+	}
+#endif
 
-	signal(SIGCHLD, (__sighandler_t)wait_child);
+	signal(SIGCHLD, (__sighandler_t)sig_chld_handler);
 	rv = loop(lock_fd);
 
 	return rv;
@@ -1534,7 +1585,8 @@ static int do_attr(void)
 	if (!cl.attr_msg.attr.tkt_id[0]) {
 		/* If the loaded configuration has only a single ticket defined, use that. */
 		if (booth_conf->ticket_count == 1) {
-			strcpy(cl.attr_msg.attr.tkt_id, booth_conf->ticket[0].name);
+			strncpy(cl.attr_msg.attr.tkt_id, booth_conf->ticket[0].name,
+				sizeof(cl.attr_msg.attr.tkt_id));
 		} else {
 			rv = 1;
 			log_error("No ticket given.");
@@ -1561,7 +1613,10 @@ out:
 int main(int argc, char *argv[], char *envp[])
 {
 	int rv;
-	char *cp;
+	const char *cp;
+#ifdef LOGGING_LIBQB
+	enum qb_log_target_slot i;
+#endif
 
 	init_set_proc_title(argc, argv, envp);
 	get_time(&start_time);
@@ -1572,15 +1627,25 @@ int main(int argc, char *argv[], char *envp[])
 	cl.lockfile[0] = 0;
 	debug_level = 0;
 
-	cl_log_set_entity(
-		(cp = strstr(argv[0], ATTR_PROG)) && !strcmp(cp, ATTR_PROG)
+
+	cp = ((cp = strstr(argv[0], ATTR_PROG)) && !strcmp(cp, ATTR_PROG)
 		? ATTR_PROG
-		: "booth"
-		);
+		: "booth");
+#ifndef LOGGING_LIBQB
+	cl_log_set_entity(cp);
+#else
+	qb_log_init(cp, LOG_USER, LOG_DEBUG);  /* prio driven by debug_level */
+	for (i = QB_LOG_TARGET_START; i < QB_LOG_TARGET_MAX; i++) {
+		if (i == QB_LOG_SYSLOG || i == QB_LOG_BLACKBOX)
+			continue;
+		qb_log_format_set(i, "%t %H %N: [%P]: %p: %b");
+	}
+	(void) qb_log_filter_ctl(QB_LOG_STDERR, QB_LOG_FILTER_ADD,
+	                         QB_LOG_FILTER_FILE, "*", LOG_DEBUG);
+#endif
 	cl_log_enable_stderr(TRUE);
 	cl_log_set_facility(0);
 
-
 	rv = read_arguments(argc, argv);
 	if (rv < 0)
 		goto out;
@@ -1607,6 +1672,9 @@ int main(int argc, char *argv[], char *envp[])
 	}
 
 out:
+#ifdef LOGGING_LIBQB
+	qb_log_fini();
+#endif
 	/* Normalize values. 0x100 would be seen as "OK" by waitpid(). */
 	return (rv >= 0 && rv < 0x70) ? rv : 1;
 }
